print("import: 'ctapipe'")
import ctapipe

print("import: 'ctapipe.io'")
import ctapipe.io

print("import: 'ctapipe.core'")
import ctapipe.core

print("import: 'ctapipe.coordinates'")
import ctapipe.coordinates

print("import: 'ctapipe.instrument'")
import ctapipe.instrument

print("import: 'ctapipe.visualization'")
import ctapipe.visualization

print("import: 'ctapipe.tools'")
import ctapipe.tools

print("import: 'ctapipe.utils'")
import ctapipe.utils

print("import: 'ctapipe.reco'")
import ctapipe.reco

